# -*- coding: utf-8 -*-
"""
Created on Sun Dec 18 10:08:43 2022

@author: anba
"""

import pandas as pd
import numpy as np
import os,sys

from scipy.ndimage import gaussian_filter1d
import argparse
import matplotlib.cm as cm
import cmocean


import vispy 
from vispy.app import use_app, Timer
from vispy import app, visuals, scene
from vispy.io import imread, load_data_file, read_mesh, read_png
from vispy.visuals.transforms import STTransform,MatrixTransform
from vispy.io import read_mesh, load_data_file, imread
from vispy.scene.visuals import Mesh
from vispy.scene import transforms
from vispy.visuals.filters import ShadingFilter, WireframeFilter,TextureFilter


from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import * 
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from scipy.ndimage import gaussian_filter1d

#%%
# from workhorse_adcp import workhorse_adcp as wh_adcp 
# from seabird_ctd import seabird_ctd as sb_ctd 

from ..ptools import ptools


import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd, numpy as np



#%%#############################################################################
#######################  INITIALIZE VISUALIZATION  ############################
###############################################################################
class PCV:
    def __init__(self):
        self.update_count = 0
        
    

        
    def init_plot(self,canvas):
        
        
        
        path = r'C:\Users\anba\OneDrive - DHI\Desktop\Visualization\PCV\PCV_Tracks_with_Time.csv'
        PCV_df = pd.read_csv(path, index_col = 0, parse_dates = True)
        PCV_df['z'].loc[PCV_df['z']>-4100] = np.mean(PCV_df['z'])
        PCV_df['z'] = PCV_df['z']
        PCV_df.drop(columns = 'id',inplace = True)
        PCV_df.dropna(inplace = True)
        
        pos = np.array([PCV_df['x'].tolist(),PCV_df['y'].tolist(),PCV_df['z'].tolist()]).T
        #PCV_carray = vispy.color.color_array.ColorArray(color=PCV_cmap(np.ones(len(PCV_df))), alpha=.1, clip=True, color_space='rgb')
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)

        self.PCV_tracks = scene.visuals.Markers(pos = pos, size = 1,parent =canvas.main_view.scene)     



        # self.ax = scene.visuals.XYZAxis(pos = self.adcp.geometry.position[:,0], parent=canvas.main_view.scene)
                                   
                                   

        
    def update(self):
        
        None

        
        
class IP_ROV:
    # y is the forward direction
    def __init__(self,name,adcp):
        # contains data 
        self.name = name # name of the asset 
        self.adcp = adcp
        
        self.platform_offset = np.array([-0.36,-1.45,-0.31]) # position of the center of the instrument platform (e.g., vessel) relative to the position measurement. 
        
        self.platform_width = 1.74
        self.platform_height = 2.74
       
        self.adcp_absolute_beam_midpoint_positions = adcp.geometry.get_absolute_beam_midpoint_positions()
        self.beam_data = adcp.get_ensemble_array(field_name = 'ABSOLUTE BACKSCATTER')
        
        
        self.timesteps = adcp.get_ensemble_datetimes()

    def init_plot(self,canvas):
        
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)
        
        
        
        self.plane_verts = np.array([[-.5,-.5,0],
                                    [-.5,.5,0],
                                    [.5,.5,0],
                                    [.5,-.5,0],
                                    [-.5,-.5,0]])
        
        Ts = np.diag([1.74,2.74,1]) #ROV x-y dimensions
        self.plane_verts = self.plane_verts.dot(Ts) - self.platform_offset
        #self.plane = scene.visuals.Line(pos = self.plane_verts, width = 5, parent  = canvas.main_view.scene)
        #self.plane = scene.visuals.Line(pos = np.add(self.plane_verts , self.adcp.geometry.position[:,0]), width = 5, parent  = canvas.main_view.scene)
        
        
        cmap = mpl.cm.jet
        c = cmap(self.beam_data.flatten()).reshape(list(np.shape(self.beam_data))+[4])
        self.beam_colors = np.full([self.adcp.config.number_of_beams,self.adcp.n_ensembles],
                          vispy.color.color_array.ColorArray)
        for b in range(self.adcp.config.number_of_beams):
            for e in range(self.adcp.n_ensembles):
                self.beam_colors[b,e] = vispy.color.color_array.ColorArray(color= c[b,:,e,:], alpha=1, clip=True, color_space='rgb')
                

        self.beam_verts = [] # store relative positions of the beams including origin
        self.beam_lines = []
        self.beam_labels = []
        for i in range(self.adcp.config.number_of_beams):
            origin = self.adcp.geometry.relative_beam_origin[:,i] 
            #pos = np.add(self.adcp.beam_midpoint_positions[i].T,self.adcp.geometry.position[:,0])
            
            pos = np.zeros((self.adcp.config.number_of_cells+1,3), dtype = float)
            pos[1:,:] = self.adcp.geometry.relative_beam_midpoint_positions[i].T
            pos[0,:] = origin
            self.beam_verts.append(pos)
           # pos = np.add(pos,self.adcp.geometry.position[:,0])

            
            self.beam_lines.append(scene.visuals.Line(pos = self.adcp_absolute_beam_midpoint_positions[i][:,:,0].T,
                                                width = 5,
                                                color = self.beam_colors[i,0],
                                                parent =canvas.main_view.scene))   
            
            
            
            
            self.beam_labels.append(scene.visuals.Text(f'beam {i+1}',
                                                      color='white',
                                                      rotation=0,
                                                      font_size = 1500,
                                                      pos = np.add(pos,self.adcp.geometry.position[:,0])[-1],
                                                      parent=canvas.main_view.scene) )
            
        self.beam_origin = scene.visuals.Markers(pos = np.add(self.adcp.geometry.position[:,0],self.adcp.geometry.relative_beam_origin.T), size = 3,parent =canvas.main_view.scene)     
        self.asset_label  = scene.visuals.Text(self.name,
                                                  color='white',
                                                  rotation=0,
                                                  font_size = 500,
                                                  pos = self.adcp.geometry.position[:,0]- self.platform_offset,
                                                  parent=canvas.main_view.scene) 
        
        
        self.ax = scene.visuals.XYZAxis(parent=canvas.main_view.scene)
        #print(type(self.ax.pos))
        
        self.ref_verts = self.ax.pos # store unit verticies
        
        
        self.ax.set_data(pos = 5*self.ref_verts + self.adcp.geometry.position[:,0] )


        # self.ax = scene.visuals.XYZAxis(pos = self.adcp.geometry.position[:,0], parent=canvas.main_view.scene)
                                   
                                   

        
    def update(self,canvas):
        t,curr_time = sim_params.get_sim_params()    
        
        
        #print(t,curr_time)
        t = np.argmin(abs(curr_time - self.timesteps))
        
        #print(t,self.timesteps[t])
        ## find the index of the closest time
        
        
        
        
        # self.asset_label.transform = STTransform(scale=(1, 1, 1), translate=self.adcp.geometry.position[:,t]-self.adcp.geometry.position[:,t-1])
        
        
        ## Build the rotation matrix
        yaw   = self.adcp.geometry.orientation[2][t]
        pitch = self.adcp.geometry.orientation[0][t]
        roll  = self.adcp.geometry.orientation[1][t]
        R = np.dot(ptools.gen_rot_x(roll),ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))
        
    
        # update the platform (ROV)
        X = np.add(self.plane_verts.dot(R), self.adcp.geometry.position[:,t])
        self.plane.set_data(pos= X, width = 1)
        
        
        ## update the beams
        for b in range(self.adcp.config.number_of_beams):
            self.beam_lines[b].set_data(pos = self.adcp_absolute_beam_midpoint_positions[b][:,:,t].T,
                                        color = self.beam_colors[b,t])
            self.beam_labels[b].pos = self.adcp_absolute_beam_midpoint_positions[b][:,:,t].T[-1]
        
        # self.asset_label.pos = self.adcp_absolute_beam_midpoint_positions[b][:,:,t].T[-1] - self.platform_offset
        # # self.ax.pos = self.adcp.geometry.position[:,t]
        
        # self.beam_origin.set_data(pos = np.add(self.adcp.geometry.position[:,t],self.adcp.geometry.relative_beam_origin.T) )


        if t>0 and t<len(self.timesteps):
            canvas.main_view.camera.center = self.adcp.geometry.position[:,t]
            #print(canvas.main_view.camera.center)
        ## function to update the Asset

class Fixed_ADCP:
    # y is the forward direction
    def __init__(self,name,adcp):
        # contains data 
        self.name = name # name of the asset 
        self.adcp = adcp
        
        self.platform_offset = np.array([0,0,0]) # position of the center of the instrument platform (e.g., vessel) relative to the position measurement. 
        
        self.platform_width = 1
        self.platform_height = 0
        self.platform_length = 1
       
        self.adcp_absolute_beam_midpoint_positions = adcp.geometry.get_absolute_beam_midpoint_positions()
        
        u,v,z,du,dv,dz,errv = adcp.get_velocity()
        self.du = du
        self.dv = dv
        self.dz = dz
        
        
        self.beam_data = adcp.get_ensemble_array(field_name = 'ABSOLUTE BACKSCATTER')
        

         
    def init_plot(self,canvas):
        
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)
        
        ## reference axis
        self.ax = scene.visuals.XYZAxis(parent=canvas.main_view.scene)
        self.ref_verts = self.ax.pos # store unit verticies
        self.ax.set_data(pos = 5*self.ref_verts + self.adcp.geometry.position[:,0] + np.array((-1,-1,0)) )
        
        
        
        ## platform reference plane
        self.plane_verts = np.array([[-.5,-.5,0],
                                    [-.5,.5,0],
                                    [.5,.5,0],
                                    [.5,-.5,0],
                                    [-.5,-.5,0]])
        
        Ts = np.diag([self.platform_width,self.platform_length,self.platform_height]) #ROV x-y dimensions
        self.plane_verts = self.plane_verts.dot(Ts) - self.platform_offset
        self.plane = scene.visuals.Line(pos = np.add(self.plane_verts , self.adcp.geometry.position[:,0]), width = 5, parent  = canvas.main_view.scene)
        
        ## platform label
        self.asset_label  = scene.visuals.Text(self.name,
                                                  color='white',
                                                  rotation=0,
                                                  font_size = 500,
                                                  pos = self.adcp.geometry.position[:,0]- self.platform_offset,
                                                  parent=canvas.main_view.scene) 
        
        
        cmap = mpl.cm.jet
        c = cmap(self.beam_data.flatten()).reshape(list(np.shape(self.beam_data))+[4])
        self.beam_colors = np.full([self.adcp.config.number_of_beams,self.adcp.n_ensembles],
                          vispy.color.color_array.ColorArray)
        for b in range(self.adcp.config.number_of_beams):
            for e in range(self.adcp.n_ensembles):
                self.beam_colors[b,e] = vispy.color.color_array.ColorArray(color= c[b,:,e,:], alpha=1, clip=True, color_space='rgb')
                
                
        
        ## ADCP Beams
        self.beam_verts = [] # store relative positions of the beams including origin
        self.beam_lines = []
        self.beam_labels = []
        for i in range(self.adcp.config.number_of_beams):
            origin = self.adcp.geometry.relative_beam_origin[:,i] 
            pos = np.zeros((self.adcp.config.number_of_cells+1,3), dtype = float)
            pos[1:,:] = self.adcp.geometry.relative_beam_midpoint_positions[i].T
            pos[0,:] = origin
            cmap = mpl.cm.bone_r
            

            self.beam_verts.append(pos)
            self.beam_lines.append(scene.visuals.Line(pos = self.adcp_absolute_beam_midpoint_positions[i][:,:,0].T,
                                                width = 5,
                                                color = self.beam_colors[i][0],
                                                parent =canvas.main_view.scene))   
            self.beam_labels.append(scene.visuals.Text(f'beam {i+1}',
                                                      color='white',
                                                      rotation=0,
                                                      font_size = 1500,
                                                      pos = np.add(pos,self.adcp.geometry.position[:,0])[-1],
                                                      parent=canvas.main_view.scene) )
        ## beam origin points    
        self.beam_origin = scene.visuals.Markers(pos = np.add(self.adcp.geometry.position[:,0],self.adcp.geometry.relative_beam_origin.T), size = 3,parent =canvas.main_view.scene)     
        
        

        
        ## progressive vector plots 
        self.progv = []
        self.start_bin = 1
        self.end_bin = 12#self.adcp.config.number_of_cells
        pos = np.add(self.calculate_progv(0,10,self.start_bin,self.end_bin).T,self.adcp.geometry.position[:,0])
        
        
        for b in range(self.end_bin-self.start_bin):
            self.progv.append(scene.visuals.Line(pos = pos[:,b,:],
                                                width = 3,
                                                parent =canvas.main_view.scene))
            
            
        
            
        
        
        ## absolute backscatter 
        
        # self.beam_color_data = []
        # PCV_cmap = cmocean.cm.gray
        # for i in range(self.adcp.config.number_of_beams):
            
        #     PCV_carray = vispy.color.color_array.ColorArray(color=PCV_cmap(np.ones(len(PCV_df))), alpha=.1, clip=True, color_space='rgb')
        
        

            
        
        
        
            
            
        
        
        

      
    def calculate_progv(self,start_ens,end_ens,start_bin,end_bin):
        
        pos = np.zeros((3,end_bin-start_bin,(end_ens - start_ens)+1))
        
        # pos[0,:,1:] = gaussian_filter1d(np.nancumsum(np.flip(self.du[start_bin:end_bin,start_ens:end_ens],axis = 1),axis = 1),1)
        # pos[1,:,1:] = gaussian_filter1d(np.nancumsum(np.flip(self.dv[start_bin:end_bin,start_ens:end_ens],axis = 1),axis = 1),1)
        # pos[2,:,1:] = gaussian_filter1d(np.nancumsum(np.flip(self.dz[start_bin:end_bin,start_ens:end_ens],axis = 1),axis = 1),1)
        
        pos[0,:,1:] = np.nancumsum(np.flip(self.du[start_bin:end_bin,start_ens:end_ens],axis = 1),axis = 1)
        pos[1,:,1:] = np.nancumsum(np.flip(self.dv[start_bin:end_bin,start_ens:end_ens],axis = 1),axis = 1)
        pos[2,:,1:] = np.nancumsum(np.flip(self.dz[start_bin:end_bin,start_ens:end_ens],axis = 1),axis = 1)
        
        pos[2] = np.add(self.adcp.geometry.get_bin_midpoints()[start_bin:end_bin],pos[2].T).T
        
 
        #pos = gaussian_filter1d(pos,sigma = 1)
        return pos 

                            

        
    def update(self):
        t,curr_time = sim_params.get_sim_params()     
        
        # self.asset_label.transform = STTransform(scale=(1, 1, 1), translate=self.adcp.geometry.position[:,t]-self.adcp.geometry.position[:,t-1])
        
        
        ## Build the rotation matrix
        yaw   = self.adcp.geometry.orientation[2][t]
        pitch = self.adcp.geometry.orientation[0][t]
        roll  = self.adcp.geometry.orientation[1][t]
        R = np.dot(ptools.gen_rot_x(roll),ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))
        
    
        # update the platform (ROV)
        X = np.add(self.plane_verts.dot(R), self.adcp.geometry.position[:,t])
        self.plane.set_data(pos= X, width = 1)
        
        
        ## update the beams
        for b in range(self.adcp.config.number_of_beams):
            self.beam_lines[b].set_data(pos = self.adcp_absolute_beam_midpoint_positions[b][:,:,t].T,
                                        color = self.beam_colors[b][t],)
            self.beam_labels[b].pos = self.adcp_absolute_beam_midpoint_positions[b][:,:,t].T[-1]
        
        self.asset_label.pos = self.adcp.geometry.position[:,t]- self.platform_offset
        # self.ax.pos = self.adcp.geometry.position[:,t]
        
        self.beam_origin.set_data(pos = np.add(self.adcp.geometry.position[:,t],self.adcp.geometry.relative_beam_origin.T) )



        #self.progv.set_data(pos = np.add(self.calculate_progv(t,t+10)[:,4,:].T,self.adcp.geometry.position[:,t]))
        
        pos = np.add(self.calculate_progv(t,t+10,self.start_bin,self.end_bin).T,self.adcp.geometry.position[:,t])
        for b in range(self.end_bin-self.start_bin):
            self.progv[b].set_data(pos = pos[:,b,:])
        
        ## function to update the Asset        

        
         
         
#%%
##To Do 

# Update CanvasWrapper to accept the Asset object and have global simulation parameters
# Before plotting resample all asset data to simulation timestep         
        

class SimParams:
    def __init__(self,timesteps):
        
        self.timestep = 1
        self.timesteps = timesteps
        
        self.t = 0
        self.curr_time = self.timesteps[self.t]
    
    def advance(self):
        # advance the simulation one timestep
        self.t = (self.t+self.timestep)%(len(self.timesteps))
        self.curr_time = self.timesteps[self.t]
        
    
    def get_sim_params(self):
        return self.t,self.curr_time
        

    
class CanvasWrapper:
    def __init__(self, assets,sim_params):
        
        #% Global Simulation Parameters 

        
        
        ##
        
        self.canvas = scene.SceneCanvas(keys='interactive',
                                        title='NORI-D Data Viewer',
                                        show=True,
                                        bgcolor = 'black',
                                        size=(1800, 1000))
        
        self.grid =  self.canvas.central_widget.add_grid()
        self.main_view =  self.grid.add_view(row=0,
                                             col=0,
                                             row_span=10,
                                             col_span = 10) # main viewer window
        
        #window for displaying company logos
        self.icon_view = self.grid.add_view(row=9,
                                            col=1,
                                            col_span = 4,
                                            row_span = 1,
                                            margin = 1,
                                            padding = 1) 
        
        #window for displaying company logos
        self.time_view = self.grid.add_view(row=8,
                                            col=7,
                                            col_span = 3,
                                            row_span =2,
                                            margin = 1,
                                            padding = 1) 
        

        self.assets = assets
        
        
        
        ## Setup the plot objects 
        self.time_display = scene.visuals.Text(sim_params.timesteps[0].strftime("%d-%b-%y %H:%M:%S"),
                                          color='white',
                                          rotation=0,
                                          font_size = 15,
                                          pos = (380,130),
                                          parent=self.time_view.scene) 
        


        for asset in self.assets:
            asset.init_plot(self)
            
        self.main_view.camera = "turntable" #'arcball' 
    
        # self.main_view.camera.zoom(2) 

        
        





class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, canvas_wrapper: CanvasWrapper, *args, **kwargs):
        super().__init__(*args, **kwargs)

        central_widget = QtWidgets.QWidget()
        main_layout = QtWidgets.QHBoxLayout()

        self.setWindowTitle("SET TITLE")
        # self.setWidth(1800)
        # self.setFixedHeight(1000)

        self.setGeometry(25, 25, 1800, 1000)

        
        self._canvas_wrapper = canvas_wrapper
        main_layout.addWidget(self._canvas_wrapper.canvas.native)
        
        central_widget.setLayout(main_layout)
        self.setCentralWidget(central_widget)   
        
        
        quit = QtWidgets.QAction("Quit", self)
        quit.triggered.connect(self.closeEvent)
        
        self.paused = False # paused indicator variable 

        # creating a QWidget object
        widget = QtWidgets.QWidget(self)
 
        # creating a vertical box layout
        layout = QtWidgets.QVBoxLayout(self)
 
        # push button 1
        self.PausePlayButton = QtWidgets.QPushButton("Pause", self)
        self.PausePlayButton.clicked.connect(self.pause_play)
        #self.StartButton.valueChanged[int].connect(self.pause_play)
        
        # time slider 
        self.TimeSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.TimeSlider.setRange(0,len(sim_params.timesteps))
        self.TimeSlider.valueChanged[int].connect(self.update_time)
        
        # timestep dial
        self.TimestepDial = QtWidgets.QDial()
        self.TimestepDial.valueChanged.connect(self.update_timestep)
        self.TimestepDial.setRange(-10,10)
        self.TimestepDial.setNotchesVisible(True)
        self.TimestepDial.setGeometry(220, 125, 200, 60)
        self.DialLabel = QtWidgets.QLabel(f"dt = {sim_params.timestep} min", self)
        self.DialLabel.setAlignment(Qt.AlignCenter)
        self.DialLabel.setGeometry(400, 125, 200, 60)
        
        # adding these buttons to the layout
        layout.addWidget(self.PausePlayButton)
        layout.addWidget(self.TimeSlider)
        layout.addWidget(self.DialLabel)
        layout.addWidget(self.TimestepDial)
        
        # setting the layout to the widget
        
        widget.setLayout(layout)
        widget.setAutoFillBackground(True)
        
        # Playback Dock Popout Window 
        self.PlaybackDock=QtWidgets.QDockWidget('Playback',self)
        self.PlaybackDock.setFixedSize(200, 300)
        self.PlaybackDock.setWidget(widget)        
        self.PlaybackDock.setGeometry(100, 0, 200, 30)
        self.PlaybackDock.setAllowedAreas(Qt.NoDockWidgetArea)

        
    def update_timestep(self):
        sim_params.timestep = self.TimestepDial.value()
        self.DialLabel.setText(f"dt = {sim_params.timestep} min")
        
        
    def update_time(self):
        sim_params.t = self.TimeSlider.value()
        #canvas.main_view.update()
        
    def pause_play(self):
        if not self.paused:
            timer.stop()
            self.paused = True
            self.PausePlayButton.setText('Play')
        else:
            timer.start()
            self.paused = False
            self.PausePlayButton.setText('Pause')
        
        
    def closeEvent(self, event):
        timer.stop()

        


def update(ev):
    
    sim_params.advance()
    t,curr_time = sim_params.get_sim_params()
    
    win.TimeSlider.setValue(t)
    canvas.time_display.text =  f't={t}'.ljust(10) + curr_time.strftime("%d-%b-%y %H:%M:%S")
    
    for asset in canvas.assets:
        
        #try:
        asset.update(canvas,sim_params)
        #except: print('failed update')

    
    

    
def run(assets,timesteps):

    global timer,canvas,app,win,sim_params
    
    
    sim_params = SimParams(timesteps = timesteps)
      
    app = use_app("pyqt5")
    app.create()
    
    
    canvas = CanvasWrapper(sim_params= sim_params, assets = assets)
    win = MainWindow(canvas)
    
    
    #timer = Timer('auto', connect=update, start=True, app = app)
    timer = Timer('auto' ,connect=update, start=True, app = app)
    win.destroyed.connect(timer.stop)
    
    win.show()
    app.run()    
  

#def run():


# assets =[ IP_ROV(name = 'ROV ADCP', adcp = adcp)]



# run(assets,timesteps = adcp.get_ensemble_datetimes())
    


#%%

# if __name__ == '__main__':
#     # assets =[data_visualizer.ROV_ADCP_Asset(name = 'HAIN', adcp = adcp1),
#     #           data_visualizer.ROV_ADCP_Asset(name = 'USBL', adcp = adcp1)]  
    
    
#     # assets =[data_visualizer.ROV_ADCP_Asset(name = 'HAIN', adcp = adcp1),
#     #           data_visualizer.ROV_ADCP_Asset(name = 'TEST', adcp = adcp1b)]  
    
#     assets =[data_visualizer.ADCP_Asset(name = 'USBL', adcp = adcp)]  
#     data_visualizer.run(assets = assets,timesteps = adcp.ensemble_times)





# -*- coding: utf-8 -*-
"""
Created on Tue Apr  4 14:56:11 2023

@author: anba
"""

import time
import pathos.multiprocessing as mp
import progressbar
import pickle
import cloudpickle    
import numpy as np
import pandas as pd

#%%



def moving_average(x, window_size):
    """
    Calculate the moving average of a one-dimensional array.

    Parameters:
        x (numpy.ndarray): The input array.
        window_size (int): The size of the moving average window.

    Returns:
        numpy.ndarray: The moving average array.

    Example:
        # Calculate the moving average of an array 'data' with a window size of 3
        result = moving_average(data, window_size=3)
    """
    return np.convolve(x, np.ones(window_size) / window_size, mode='same')

def moving_variance(x, window_size):
    """
    Calculate the moving variance of a one-dimensional array.

    Parameters:
        x (numpy.ndarray): The input array.
        window_size (int): The size of the moving variance window.

    Returns:
        numpy.ndarray: The moving variance array.

    Example:
        # Calculate the moving variance of an array 'data' with a window size of 3
        result = moving_variance(data, window_size=3)
    """
    mean_centered = x - np.mean(x)
    return np.convolve(mean_centered**2, np.ones(window_size) / window_size, mode='same')




def pickle_obj(obj,fname):
    """
    Pickle an object

    Parameters
    ----------
    obj : Object to pickle
    fname : string
        filename for pickled object

    Returns
    -------
    None.

    """

    
    with open(fname,"wb") as f:
        
        try:
            pickle.dump(obj,f)
        except:
            cloudpickle.dump(obj,f)
        
def unpickle_obj(fname):
    """
    Unpickle an object

    Parameters
    ----------
    fname : string
        Path to pickled object.

    Returns
    -------
    out : unpickled object.

    """

    with open(fname,"rb") as f:
        
        try:
            out = pickle.load(f) 
        except:
            out = cloudpickle.load(f) 
    return out

def flatten_list(l): # flatten a list of lists
    """
    Parameters
    ----------
    l : list
        List to flatten

    Returns
    -------
    out : flattened list

    """
    
    out = [item for sublist in l for item in sublist]
    return out

def printProgressBar (iteration, total,taskname = '', prefix = 'Progress', decimals = 1, length = 25, fill = '█', printEnd = "\r"):
    """
    Terminal progress bar
    Args:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        taskname    - Optional  : task description (Str)
        prefix      - Optional  : prefix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
   
    
    taskname = taskname.ljust(100)
    print(f'\r{prefix} |{bar}|{taskname}', end = printEnd)
    # # Print New Line on Complete
    # if iteration == total: 
    #     print()  
        
        
def paralell_fcn(f,iterable,taskname = 'Processing'):
    
    
    """
    Run a function in paralell

    Parameters
    ----------
    f : function to run 
    iterable: inpute argument to interate over
    taskname: task label on progress bar
    Returns
    -------
    list of task results

    """

    st = time.time()
    cpu_count = min(mp.cpu_count(),len(iterable))
    varname = 'Task'# '_'.join(taskname.split(' '))
    widgets = [progressbar.GranularBar(markers=" ░▒▓█", left='', right='|'),
                progressbar.widgets.Variable(varname)]
    bar = progressbar.ProgressBar(widgets = widgets, max_value = len(iterable))
    bar.variables[varname] = str(taskname) 
    bar.update(1)
    results = []
    
    #printProgressBar(0, len(iterable),prefix = f'Filtering', taskname= f'Filtering 0%')
        
    with mp.Pool(cpu_count) as pool:
        
        
        #results= pool.map(f,iterable)
        tasks = pool.imap_unordered(f,iterable) 
        
        for r,result in enumerate(tasks):
            bar.update(r+1)
            
            
            
            results.append(result)
            time.sleep(0.1)
            
            #print(r)
    bar.update(r+1)
    #bar.update(len(iterable)) # force bar to 100%
            
    # print('\r','\n')
    # print(f'Execution Time: {round((time.time() - st)/60,2)} min')
    # print('\r','\n')
    return results


def gen_rot_y(theta):
    """
    Generate a 3*3 matrix that rotates a vector by angle theta about the y-axis

    Parameters
    ----------
    theta : float
        angle to rotate by.

    Returns
    -------
    numpy array
        Rotation matrix (Ry).

    """
    theta = theta*np.pi/180
    return np.array([(np.cos(theta),0,np.sin(theta)),
                     (0,1,0),
                     (-np.sin(theta),0,np.cos(theta))])

def gen_rot_z(theta):
    """
    Generate a 3*3 matrix that rotates a vector by angle theta about the z-axis

    Parameters
    ----------
    theta : float
        angle to rotate by.

    Returns
    -------
    numpy array
        Rotation matrix (Rz).

    """
    theta = theta*np.pi/180
    return np.array([(np.cos(theta),-np.sin(theta),0),
                     (np.sin(theta),np.cos(theta),0),
                     (0,0,1)])
                     
def gen_rot_x(theta):
    """
    Generate a 3*3 matrix that rotates a vector by angle theta about the x-axis

    Parameters
    ----------
    theta : float
        angle to rotate by.

    Returns
    -------
    numpy array
        Rotation matrix (Rx).

    """
    theta = theta*np.pi/180
    return np.array([(1,0,0),
                     (0,np.cos(theta),-np.sin(theta)),
                     (0,np.sin(theta),np.cos(theta))])

def rolling_spike_filter(data, window_size=3, threshold_factor=3):
    """
    Apply a rolling median and Median Absolute Deviation (MAD) based filter to remove spikes from data.

    Parameters
    ----------
    data : array-like
        The input data from which spikes are to be removed.
    window_size : int, optional
        The size of the window used to compute the rolling median and MAD. Defaults to 3.
    threshold_factor : int, optional
        The factor used to determine the threshold for spike detection. A data point is considered a spike
        if its deviation from the rolling median exceeds this factor times the MAD. Defaults to 3.

    Returns
    -------
    numpy.ndarray
        The filtered data array with spikes removed and replaced through forward and backward filling.

    Notes
    -----
    This function is useful for time-series data where transient, large deviations (spikes) need to be removed
    to prevent distortions in further analysis or processing.
    """
    series = pd.Series(data)
    rolling_median = series.rolling(window=window_size, center=True).median()
    deviation = np.abs(series - rolling_median)
    mad = deviation.rolling(window=window_size, center=True).median()

    outliers = deviation > threshold_factor * mad
    filtered_series = series.where(~outliers, other=np.nan)
    filtered_series.fillna(method='bfill', inplace=True)
    filtered_series.fillna(method='ffill', inplace=True)

    return filtered_series.to_numpy()
###########################################################################################################
##########################################################################################################

def configure_ADCP(adcp, pose, **kwargs):
    """
    Configure the ADCP parameters, perform processing, and apply masking.

    Parameters
    ----------
    adcp : ADCP object
        An instance of the ADCP class.
    pose : Pose object
        Pose information of the ADCP.
    adcp_offset : ndarray, optional
        Offset of the ADCP from the ROV center of mass.
    adcp_rotation : float, optional
        Rotation of the ADCP in its housing.
    adcp_dr : float, optional
        Distance between transducer faces for the ADCP.
    stn_thresh : int, optional
        Threshold for Signal-to-Noise Ratio masking.
    cmag_thresh : int, optional
        Threshold for Correlation Magnitude masking.
    abs_thresh : int, optional
        Threshold for Absolute Backscatter masking.
    reject_first_bins : list, optional
        List of first n bins to reject (bin 1 - reject_first_bins).

    Returns
    -------
    None

    Notes
    -----
    - The ADCP geometry is updated based on the provided pose and optional parameters.
    - Absolute backscatter is calculated.
    - Various masks (stn, cmag, abs_bs, bins, bottom track) are defined and applied.

    Example
    -------
    configure_ADCP(adcp_instance, pose_instance, adcp_offset=my_offset, adcp_rotation=my_rotation,
                   adcp_dr=my_distance_range, stn_thresh=my_stn_threshold, cmag_thresh=my_cmag_threshold,
                   abs_thresh=my_abs_threshold, reject_first_bins=my_reject_bins)

    """
    ips_offset = np.array([-0.36, -1.45, -0.31])  # Offset of the ips from the ROV center of mass
    adcp_offset = np.array([0.23, -1.38, -0.19])  # Offset of the ADCP from the ROV center of mass
    adcp_offset -= ips_offset
    adcp_length = 0.3949
    adcp_offset[2] += adcp_length
    adcp_offset[2] = -adcp_offset[2]

    offset = kwargs.pop('adcp_offset', adcp_offset)
    adcp_rotation = kwargs.pop('adcp_rotation', 0)
    adcp_dr = kwargs.pop('adcp_dr', 0.1)
    stn_thresh = kwargs.pop('stn_thresh', None)
    cmag_thresh = kwargs.pop('cmag_thresh', None)
    abs_thresh = kwargs.pop('abs_thresh', None)
    reject_first_bins = kwargs.pop('reject_first_bins', None)

    adcp.mask.set_mask_status(False)
    adcp.geometry.set_pose(pose, update_orientation=True)
    adcp.geometry.calculate_beam_geometry(rotation=adcp_rotation, offset=offset, dr=adcp_dr)

    # Processing
    adcp.processing.calculate_absolute_backscatter()

    # Masking (common masking operations)

    if stn_thresh:
        stn = adcp.get_ensemble_array(field_name='SIGNAL TO NOISE RATIO')
        mask1 = (stn > stn_thresh)
        adcp.mask.define_mask(mask=mask1, name='stn', set_active=True)

    if cmag_thresh:
        cmag = adcp.get_ensemble_array(field_name='CORRELATION MAGNITUDE')
        mask2 = cmag > cmag_thresh
        adcp.mask.define_mask(mask=mask2, name='cmag', set_active=True)

    if abs_thresh:
        abs_bs = adcp.get_ensemble_array(field_name='ABSOLUTE BACKSCATTER')
        mask3 = abs_bs > abs_thresh
        adcp.mask.define_mask(mask=mask3, name='abs_bs', set_active=True)

    if reject_first_bins:
        mask4 = np.full(np.shape(mask3), True)
        for i, b in enumerate(reject_first_bins):
            mask4[:, :b, :] = False
            adcp.mask.define_mask(mask=mask4, name='bins', set_active=True)

    # # Mask bottom track
    # bt_range = adcp.get_bottom_track()
    # if any(~np.isnan(bt_range).flatten()):
    #     bt_range = np.repeat(bt_range[:, np.newaxis, :], adcp.config.number_of_cells, axis=1)
    #     bt_range[bt_range == 0] = 32675
    #     bin_centers = np.outer(adcp.geometry.get_bin_midpoints(), np.ones(adcp.n_ensembles))
    #     bin_centers = np.repeat(bin_centers[np.newaxis, :, :], adcp.config.number_of_beams, axis=0)
    #     mask = (bt_range - adcp.config.depth_cell_length / 100) > bin_centers
    adcp.processing.mask_bottom_track(cell_offset=0, spike_filter=True, spike_window=10, spike_thresh=5)
        
        



def configure_CTD(ctd,pose,min_turb = -1,ssc_coeff = 1.5,**kwargs):
    """
    configure a ctd sensor 

    Parameters
    ----------
    ctd : TYPE
        DESCRIPTION.
    Pose : Pose object
        pre-configured Pose object containing position and orientation data 
        for the sensor. 
    orient_in : TYPE
        DESCRIPTION.
    min_turbidity: float, optional
        minimium allowed turbidity value in masking 
        
    ssc_coeff: float, optional
        coefficient for converting Turbidity (NTU) to ssuspended solids concentration (SSC). SSC = ssc_coeff*NTU
    Returns
    -------
    None.

    """

    #min_turb = kwargs.pop('min_turb',0) # minimim allowed turbidity 
    offset = kwargs.pop('ctd_offset',(0,0,0))
    ctd_rotation = kwargs.pop('ctd_rotation',0)
    

    ## configure geometry
    ips_offset = np.array([	-0.36,-1.45,	-0.31]) # offset of the ips from the ROV center of mass - note x and y coordinates interchanged, and z direction reversed to match the ADCP reference frame, 
    ctd_offset = np.array([.80,1.05,	-0.52]) #np.array([0.23,-1.38,	-0.19]) # offset of the wetlabs ntu face from the ROV center of mass - note x and y coordinates interchanged, and z direction reversed to match the ADCP reference frame, 
    ctd_offset -= ips_offset 
    ctd_length = 0
    ctd_offset[2] += ctd_length # add the offset to the face of the transducers
    ctd_offset[2] = -ctd_offset[2] # flip z coordinate direction    



    ctd.geometry.set_pose(pose)
    ctd.geometry.set_sensor_relative_geometry(rotation = ctd_rotation,offset = ctd_offset)
    ctd.processing.calculate_density()
    ctd.processing.calculate_depth(environment = 'SALTWATER')
    ctd.processing.calculate_SSC(A =ssc_coeff)
    
    t,turb = ctd.get_timeseries_data(mask = False)
    mask = turb >= min_turb
    ctd.mask.define_mask(mask,name = 'turbidity',set_active = True)
    
def select_adcp_data(adcp, t_select='time', d_select='HAB', field_name='ECHO INTENSITY', **kwargs):
    """
    Selects data from an ADCP transect based on time or ensemble index and depth (HAB or bin).

    Parameters:
    - adcp (ADCP): ADCP DataSet object 
    - t_select (str): Selection method for time ('time' or 'ensemble').
    - d_select (str): Selection method for depth ('HAB' or 'bin').
    - field_name (str): Name of the field to select (e.g., 'ECHO INTENSITY').
    - **kwargs: Additional keyword arguments based on the selection criteria.

    Returns:
    - t (ndarray): Array of selected time or ensemble values.
    - X (ndarray): Selected data array based on the specified criteria.
    - Z (ndarray): z-values of selected ADCP ensemble array data 
    """
    if t_select == 'time':
        st = kwargs.pop('start_time', min(adcp.get_ensemble_datetimes()))
        et = kwargs.pop('end_time', max(adcp.get_ensemble_datetimes()))
    elif t_select == 'ensemble':
        se = kwargs.pop('start_ensemble', 0)
        ee = kwargs.pop('end_ensemble', adcp.n_ensembles)

    if d_select == 'HAB':
        sh = kwargs.pop('start_HAB', 0)
        eh = kwargs.pop('end_HAB', np.nanmax(adcp.geometry.get_absolute_beam_midpoint_positions_HAB()[2]))

    if d_select == 'bin':
        sb = kwargs.pop('start_bin', 0)
        eb = kwargs.pop('end_bin', adcp.config.number_of_cells)
    mask = kwargs.pop('mask', True)



    if d_select == 'HAB':
        # define a mask to nan out out of range data
        Z = adcp.geometry.get_absolute_beam_midpoint_positions_HAB()[2].copy()
        hab_mask = ((Z >= sh) & (Z <= eh))
        Z[~hab_mask] = np.nan

        adcp.mask.set_mask_status(mask)  # deactivate all existing masks if mask is set to false, otherwise keep existing masks
        adcp.mask.define_mask(mask=hab_mask, name='HAB', set_active=True)
        X = adcp.get_ensemble_array(field_name=field_name, mask=mask)
        # delete the HAB mask
        adcp.mask.delete_mask('HAB')

    if t_select == 'time':
        # find starting ensemble and ending ensemble based on time
        t = adcp.get_ensemble_datetimes()

        se = np.argmin(abs(st - t))  # starting ensemble index
        ee = np.argmin(abs(et - t))  # ending ensemble index
        X = X[:,:,se:ee]
        Z = Z[:,:,se:ee]
        t = t[se:ee]
    
        
    
    return t, X,Z

def select_ctd_data(ctd, field_name='Turbidity (NTU)', **kwargs):
    """
    Selects data from a CTD profile based on time.

    Parameters:
    - ctd (CTD): CTD DataSet object containing the data.
    - field_name (str): Name of the field to select (e.g., 'Turbidity (NTU)').
    
    - **kwargs: Additional keyword arguments based on the selection criteria.

    Returns:
    - t (ndarray): Array of selected time values.
    - turb (ndarray): Selected turbidity data array based on the specified criteria.
    """
    turb, t = ctd.get_timeseries_data(field_name=field_name)
    st = kwargs.pop('start_time', min(t))
    et = kwargs.pop('end_time', max(t))
    sidx = np.argmin(abs(st - t))  # starting ensemble index
    eidx = np.argmin(abs(et - t))  # ending ensemble index
    t = t[sidx:eidx]
    turb = turb[sidx:eidx]
    return t, turb



def calculate_time_overlap(start_time1, end_time1, start_time2, end_time2):
    """
    Calculate the overlap duration between two time intervals.

    Parameters:
    - start_time1: Start time of the first interval. Accepts pandas, numpy, or datetime datetimes.
    - end_time1: End time of the first interval. Accepts pandas, numpy, or datetime datetimes.
    - start_time2: Start time of the second interval. Accepts pandas, numpy, or datetime datetimes.
    - end_time2: End time of the second interval. Accepts pandas, numpy, or datetime datetimes.

    Returns:
    - overlap_duration (timedelta): Duration of the overlap between the two intervals.

    Example:
    >>> start_time1 = "2023-01-01 08:00:00"
    >>> end_time1 = "2023-01-01 12:00:00"
    >>> start_time2 = "2023-01-01 10:00:00"
    >>> end_time2 = "2023-01-01 14:00:00"
    >>> overlap_duration = calculate_time_overlap(start_time1, end_time1, start_time2, end_time2)
    >>> print(f"Overlap duration: {overlap_duration}")
    """
    # Convert inputs to datetime objects if they are not already
    start1 = pd.to_datetime(start_time1)
    end1 = pd.to_datetime(end_time1)
    start2 = pd.to_datetime(start_time2)
    end2 = pd.to_datetime(end_time2)

    overlap_start = max(start1, start2)
    overlap_end = min(end1, end2)

    if overlap_start < overlap_end:
        overlap_duration = overlap_end - overlap_start
        return overlap_duration
    else:
        return pd.Timedelta('0s')
    

def find_ADCP_CTD_pairs(dm, min_overlap=300, plot=True):
    """
    Find matched ADCP and CTD pairs based on overlapping date ranges.

    Parameters:
    - dm: DataManager object containing datasets named 'ADCP' and 'CTD' with ADCP and CTD DataSet objects, respectively.
    - min_overlap: Minimum number of seconds two files must overlap to be considered a match.
    - plot: bool - if True, plot the matched pairs.

    Returns:
    - list: List of tuples containing matched ADCP and CTD pairs (adcp, ctd).
    """
    pairs = []

    for ctd in dm.datasets['CTD'].get_active_data():
        for adcp in dm.datasets['ADCP'].get_active_data():
            
            # Check for valid file lengths and dates
            fail_conds = []  # Boolean fail indicators - for rejecting a CTD or ADCP dataset
            fail_conds.append(ctd.data['DateTime'].min() < pd.to_datetime('September 01 2022'))
            fail_conds.append((ctd.data['DateTime'].max() - ctd.data['DateTime'].min()) < pd.Timedelta('5 min'))
            fail_conds.append((adcp.get_ensemble_datetimes().max() - adcp.get_ensemble_datetimes().min()) < pd.Timedelta('5 min'))

            if not any(fail_conds): 
                
                overlap = calculate_time_overlap(ctd.data['DateTime'].min(), ctd.data['DateTime'].max(), adcp.get_ensemble_datetimes().min(), adcp.get_ensemble_datetimes().max())
                if overlap.total_seconds() > min_overlap:
                    pairs.append((adcp, ctd))
                    # You can perform additional actions here for matched pairs, such as plotting
                    
                    if plot:
                        adcp.plot.four_beam_flood_plot(plot_by='bin', ctd=ctd)

    return pairs


def ABS_to_NTU(ABS, A, B):
    '''
    Convert ABS to NTU using the equation 10log(SSC) = A*Backscatter + B.

    Parameters
    ----------
    ABS : numpy.ndarray
        Absolute backscatter values
    A : float
        Coefficient A in the equation.
    B : float
        Coefficient B in the equation.

    Returns
    -------
    numpy.ndarray
        Nephelometric Turbidity Units (NTU) calculated from ABS.
    '''

    ABS[ABS >= 32768] = np.nan
    return (1 / 10) * 10**(A * ABS + B) 

def NTU_to_SSC(NTU,A):
    '''
    Convert NTU to SSC using equation SCC = A*NTU

    Parameters
    ----------
    NTU : numpy.ndarray
        Absorption values.
    A : float
        Coefficient A in the equation.


    Returns
    -------
    numpy.ndarray
        Suspended solids concentration (SSC) calcualted from NTU
    '''
    return A*NTU  

def random_sample(array, n):
    """
    Randomly sample n elements from the given array.

    Parameters:
    - array: Input array.
    - n: Number of elements to sample.

    Returns:
    - np.ndarray: Array containing n randomly sampled elements.
    """
    if n > len(array):
        raise ValueError("Sample size cannot be greater than the array size.")
    
    return np.random.choice(array, size=n, replace=False)


def linear_regression(X, y):
    """
    Computes the coefficients for a linear regression using the normal equations,
    handling NaN values by replacing them with the mean of the respective column.
    
    Args:
    X : numpy.ndarray
        The matrix of input features. Each row represents a sample,
        and each column represents a feature.
    y : numpy.ndarray
        The vector of target values corresponding to each row in X.
    
    Returns:
    numpy.ndarray
        The vector of computed coefficients.
    """
    # Handling NaNs: replace NaNs in X with the column mean
    col_means = np.nanmean(X, axis=0)
    # Find indices where NaNs are located
    inds = np.where(np.isnan(X))
    # Place column means in the positions of NaNs
    X[inds] = np.take(col_means, inds[1])

    # Handling NaNs in y: replace NaNs with the mean of y
    y_mean = np.nanmean(y)
    y = np.where(np.isnan(y), y_mean, y)
    
    # Add a column of ones to X to account for the intercept term
    X_b = np.c_[np.ones((X.shape[0], 1)), X]
    
    # Compute the matrix product of X_b's transpose and X_b
    X_b_T_X_b = X_b.T.dot(X_b)
    
    # Compute the inverse of X_b_T_X_b
    X_b_T_X_b_inv = np.linalg.inv(X_b_T_X_b)
    
    # Compute the matrix product of the inverse of X_b_T_X_b and the transpose of X_b
    X_b_T_X_b_inv_X_b_T = X_b_T_X_b_inv.dot(X_b.T)
    
    # Compute the final coefficients by multiplying with y
    theta = X_b_T_X_b_inv_X_b_T.dot(y)
    
    return theta


import pyplume.plotting.matplotlib_shell as plotting
import pyplume.ptools.ptools as ptools
import pyplume.pd0.pd0_decoder as pd0
import pyplume.ctd.seabird_hex_decoder as ctd
import pyplume.pose.Pose as pose
import pyplume.data_manager.data_manager as data_manager
import pyplume.data_visualizer.visualizer as visualizer
import pyplume.data_visualizer.asset_templates as templates




# -*- coding: utf-8 -*-
"""
Created on Wed Aug  2 16:34:15 2023

@author: anba
"""

class PT3:
    # class to manage PT3 teste results. Only parsing High Gain RSSI implemented
    # currently.
    def __init__(self,filepath):
        self.filepath = filepath
        self.k_c = self.read()
    def read(self):
        """
        Read the High Gain RSSI parameters from a TRDI PT3 calibration test file

        Parameters
        ----------
        path : string
            path to the .txt file containing PT3 test data

        Returns
        -------
        k_c : dict
            dictionary containign the parsed RSSI parameters for each beam 

        """
        file = open(self.filepath,'r')
        for i,line in enumerate(file.readlines()):
            L = [i for i in line.split(' ') if i]
        
            if ' '.join(L[:3]) == 'High Gain RSSI:':
                k_c = { 1: float(L[3])/100,# beam 1
                        2: float(L[4])/100,# beam 2
                        3: float(L[5])/100,# beam3
                        4: float(L[6].replace('\n',''))/100}# beam4  
        file.close()
        return k_c


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyplume-dhi",
    version="0.0.31",
    author="Andy Banks",
    author_email="anba@dhigroup.com",
    description="A library for reading, manipulating, and analyzing TRDI ADCP data and SeaBird CTD data.", 
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=" ",
    packages=['pyplume','pyplume.pd0','pyplume.ctd','pyplume.plotting','pyplume.ptools','pyplume.pose','pyplume.data_manager','pyplume.data_visualizer'],#['pyplume','pyplume.pd0','pyplume.ctd','pyplume.plotting','pyplume.ptools','pyplume.data_manager','pyplume.data_visualizer','pyplume.pose']
    install_requires=[
        'numpy',
        'rich',
        'pandas',
        'matplotlib',
        'seaborn',
        'cmocean',
        'pyproj',
        'dill',
        'pathos',
        'progressbar',
        'cloudpickle',
	'psutil',
	'scipy',
    ]   ,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




# -*- coding: utf-8 -*-
"""
Created on Sat Jul 22 08:55:44 2023

@author: anba
"""
import numpy as np 

class masking:
    def __init__(self,pd0,**kwargs):
        
        self._pd0 = pd0
        
        self.masks = {} # dict to hold masks for ensemble data
        self.mask_status= {} # dict to hold activation status of defined masks 
        
        
        
    def delete_mask(self,name):
        """
        Define a mask
    
        Parameters
        ----------

        name : str
            name for the mask

        Returns
        -------
        None.
    
        """     
        
        
        try:
            del self.masks[name]
            del self.mask_status[name]
            #print('delete successful')
        except:
            ValueError(f'Invalid mask name')
            
            
        
        
        
            
    def define_mask(self,mask,name, set_active = False):
        """
        Define a boolean mask that can be applied to ensemble data when ensemble
        data is called. The ensemble data corresponding FALSE entries in the mask
        will be converted to np.nan 
    
        Parameters
        ----------
        mask : numpy array
            numpy array with dimensions (n_beams,n_bins,n_ensembles).
        name : str
            name for the mask
        set_active: bool
            if True, the defined mask will be activated. All calls to 
            get_ensemble_array will have this mask applied. Default is False.  
    
        Returns
        -------
        None.
    
        """
        #check that mask has correct dimensions  
        valid_dims = (self._pd0.config.number_of_beams,self._pd0.config.number_of_cells,self._pd0.n_ensembles)
        if not np.shape(mask) == valid_dims:
            ValueError(f'Mask must have dimensions of {valid_dims}')
        
    
        
        self.masks[name] = mask
        
    
        # initalize the mask status 
        if set_active:
            self.mask_status[name] = True
        else:
            self.mask_status[name] = False
      

    def set_mask_status(self,status,name = 'all'):
        """
        Change the status of a mask, or multiple masks.The ensemble data 
        corresponding to FALSE entries in the mask will be converted to np.nan

        Parameters
        ----------
        name : str or list of str, optional
            mask name, or list of mask names. if 'all', all defined masks will be 
            activated. The default is 'all'.
        status : TYPE
            DESCRIPTION.

        Raises
        ------
        ValueError
            Invalid mask name

        Returns
        -------
        None.

        """
        
        
        if name == 'all':
            for key in self.masks.keys():
                self.mask_status[key] = status
                
        elif type(name) == list:
            for n in names:
                # check if mask exists 
                if n in self.masks.keys(): 
                    self.mask_status[name] = status
                else: 
                    raise ValueError(f'{name} is not a defined mask.')
                
        elif type(name) == str:
            if name in self.masks.keys(): 
                self.mask_status[name] = status
            else: 
                raise ValueError(f'{name} is not a defined mask.')
        
        
        
    def apply_masks(self,X):
        """
        apply all of the active masks to the ensemble array X. Masked values 
        will be converted to np.nan
        

        Parameters
        ----------
        X : numpy array
            array of ensemble data with dimensions (n_bins,n_ensembles).

        Returns
        -------
        X : numpy array
            masked array of ensemble data with dimensions (n_bins,n_ensembles).

        """
        #combined_mask = np.full((self._pd0.config.number_of_cells,self.n_ensembles),False)
        if self.masks:
            for mask_name in self.masks.keys():
                if self.mask_status[mask_name]:
                    mask = self.masks[mask_name]
                    
                    #print(np.shape(mask), np.shape(X))
                    np.putmask(X,~mask, values = np.nan)


        return X
        
        


# -*- coding: utf-8 -*-
from datetime import datetime
import cycler
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.patheffects as mpl_path_effects

#class plotting:
    
dhi_blue1 = '#04426e'
dhi_blue2 = '#4d9ab3'
dhi_blue3 = '#0493b2'
dhi_blue4 = '#c3dde5'
dhi_green1 = '#01be62'
dhi_green2 = '#00b591'
dhi_green3 = '#6ad6af'


dhi_gray1 = '#c4c4c4'
dhi_gray2 = '#8b8b8c'
dhi_gray3 = '#686c6e'

dhi_red1= '#c81f00'
dhi_red2 = '#ac1817'

dhi_yellow1 = '#ffbb3c'
dhi_yellow2 = '#ebd844'

dhi_orange1 = '#ec8833'
dhi_orange2 = '#d3741c'


class dhi_colors:
    blue1 = '#04426e'
    blue2 = '#4d9ab3'
    blue3 = '#0493b2'
    blue4 = '#c3dde5'
    
    green1 = '#93c47d' #'#01be62'
    green2 = '#00b591'
    green3 = '#6ad6af'
    
    gray1 = '#c4c4c4'
    gray2 = '#8b8b8c'
    gray3 = '#686c6e'

    red1= '#c81f00'
    red2 = '#ac1817'

    yellow1 = '#ffbb3c'
    dhi_yellow2 = '#ebd844'

    orange1 = '#ec8833'
    orange2 = '#d3741c'
    
    
    
    




mpl.rcParams['font.size'] = 9
mpl.rcParams['lines.linewidth'] = 2
mpl.rcParams['lines.color'] = 'black'
#mpl.rcParams['axes.grid'] = True
mpl.rcParams['patch.edgecolor'] = 'white'
mpl.rcParams['axes.grid.which'] = 'major'
mpl.rcParams['lines.markersize'] = 1.6
mpl.rcParams['ytick.labelsize'] = 8
mpl.rcParams['xtick.labelsize'] = 8

mpl.rcParams['ytick.labelright'] = False
mpl.rcParams['xtick.labeltop'] = False

mpl.rcParams['ytick.right'] = True
mpl.rcParams['xtick.top'] = True
mpl.rcParams['ytick.major.right'] = True
mpl.rcParams['xtick.major.top'] = True


# #mpl.rcParams['text.latex.preamble']=[r"\usepackage{lmodern}"]

# params = {'text.usetex' : True,
#           'font.size' : 11,
#           'font.family' : 'lmodern',
#           'text.latex.unicode': True,
#           }
# mpl.rcParams.update(params) 

# mpl.rcParams['axes.titlepad']: 250
# mpl.rcParams['axes.labelpad']: 250
mpl.rcParams['axes.labelweight'] = 'normal'

mpl.rcParams['legend.fontsize'] = 8
mpl.rcParams['legend.framealpha']= 0.5

mpl.rcParams['axes.titlesize'] = 12
mpl.rcParams['axes.titleweight'] ='normal'

mpl.rcParams['font.family'] ='TImes New Roman'
#rcParams['font.family'] = 'Georgia'

mpl.rcParams['axes.labelsize'] = 10
mpl.rcParams['axes.linewidth'] = 1.25

mpl.rcParams['xtick.major.size'] = 5.0
mpl.rcParams['xtick.minor.size'] = 3.0
mpl.rcParams['ytick.major.size'] = 5.0
mpl.rcParams['ytick.minor.size'] = 3.0


mpl.rcParams['figure.dpi'] : 300.0



# blues  = ['#9ecae1','#6baed6','#4292c6','#2171b5']#,'#08519c']
# greens = ['#c7e9c0','#a1d99b','#74c476','#41ab5d']#,'#238b45']
# colors = []
# for i in range(len(blues)):
#     colors.append(blues[-(i+1)])
#     colors.append(greens[-(i+1)])

colors = 2*['#283747','#0051a2', '#41ab5d', '#feb24c', '#93003a']
line_style = 5*['-'] + 5*['--']
mpl.rcParams['axes.prop_cycle'] = cycler.cycler('color',colors) +cycler.cycler('linestyle',line_style)

alpha = 0.7
to_rgba = mpl.colors.ColorConverter().to_rgba#
color_list=[]
for i, col in enumerate(mpl.rcParams['axes.prop_cycle']):
    color_list.append(to_rgba(col['color'], alpha))
mpl.rcParams['axes.prop_cycle'] = cycler.cycler(color=color_list)


mpl.rcParams['xtick.direction'] = 'in'
mpl.rcParams['ytick.direction'] = 'in'


    

def subplots(**kwargs):
    
    if kwargs.get('figheight'): figheight = kwargs.get('figheight')
    else: figheight = 4.25*(1+(5**.5))/2
    
    if kwargs.get('figwidth'): figwidth = kwargs.get('figwidth')
    else: figwidth = figheight
    
    
    if kwargs.get('nrow'): nrow = kwargs.get('nrow')
    else: nrow = 1
    
    if kwargs.get('ncol'): ncol = kwargs.get('ncol')
    else: ncol = 1    
    
    if kwargs.get('sharex'): sharex = kwargs.get('sharex')
    else: sharex = False   
    
    if kwargs.get('sharey'): sharey = kwargs.get('sharey')
    else: sharey = False 
    
    if kwargs.get('width_ratios'): width_ratios = kwargs.get('width_ratios')
    else: width_ratios = [1]*ncol

    if kwargs.get('height_ratios'): height_ratios = kwargs.get('height_ratios')
    else: height_ratios = [1]*nrow   
    
    
    if kwargs.get('watermark_text') : watermark_text = kwargs.get('watermark_text')
    else: watermark_text = 'PRELIMINARY RESULTS \n INTERNAL CIRCULATION ONLY'
    
    
    
    fig,axs = plt.subplots(figsize = (figwidth,figheight),
                           nrows = nrow,
                           ncols = ncol,
                           gridspec_kw={'width_ratios': width_ratios,
                                        'height_ratios': height_ratios},
                           sharex = sharex,
                           sharey = sharey)


  
    def add_watermark(ax, add_date = True):
        # add a watermark and some metadata to the axes
        path_effects=[mpl_path_effects.Stroke(linewidth=2, foreground="black",alpha = 0.035)] 
        ax.text(.5, .5, watermark_text, transform=ax.transAxes,fontsize=15, color='gray', alpha=.1,ha='center', va='center', rotation=0,path_effects = path_effects,fontname = 'Arial',zorder = 1)
        if add_date:
            ax.text(.985, 0.02,os.getlogin() + '\n' + datetime.now().strftime("%d/%m/%y %H:%M"), transform=ax.transAxes,fontsize=6, color='black', alpha=0.4,ha='right', va='bottom', rotation=0, weight = 'bold',fontname = r'Arial',zorder = 1)
        return ax
    
    
 
    def set_ax_params(ax):
        
            # if not kwargs.get('for_production'):
            #     ax = add_watermark(ax)
            # else: ax = add_watermark(axs)
            
            # ax.xaxis.set_major_locator(mpl.ticker.AutoLocator())
            # ax.yaxis.set_major_locator(mpl.ticker.AutoLocator())
            
            # ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())
            # ax.yaxis.set_minor_locator(mpl.ticker.AutoMinorLocator())

            ax.grid(alpha = 0.25)
        
    
    if nrow*ncol>1:
        
        for i,ax in enumerate(axs.reshape(-1)): 
            set_ax_params(ax)
            
    else:
        set_ax_params(axs)
            

        
                

             
       

    
    

                    
 

    plt.tight_layout(rect = [0.025,0.1,0.975,.9])
    return fig,axs








# -*- coding: utf-8 -*-
"""
Created on Fri Jul 21 15:45:10 2023

@author: anba
"""
import numpy as np
from scipy.ndimage import gaussian_filter
from ..ptools import ptools
class processing:
    
    def __init__(self,pd0):
        
        self._pd0 = pd0
        
        self.ensemble_fields = ['ECHO INTENSITY','CORRELATION MAGNITUDE','PERCENT GOOD'] # keys for fields defined in ensemble data. Can be modified by append_to_ensembles method
        
        
        #vectorize some functions
        self.water_absorption_coeff = np.vectorize(self._scalar_water_absorption_coeff)
        self.sediment_absorption_coeff = np.vectorize(self._scalar_sediment_absorption_coeff)   
        self.counts_to_absolute_backscatter = np.vectorize(self._scalar_counts_to_absolute_backscatter)


    def apply_interference_filter(self,niter = 3, dx_thresh = 0.05, sigma = None):
        
        def filter_array(X):
            """
            Generate numpy array of filterd Data (designed to remove HAIN DVL interference)
                -Two passes of column-wise spike filter, filling nan values inbetween passes by averaging neighbors.
                -Final gaussing smoothing step
            Args:
                ensemble_array: numpy array of echo intensity data with dimensions (nbins,n_ensembles)
            Returns:
                numpy array with dimensions (nbins,n_ensembles)  
        
            """     
            
            def filt(X):
                """
                column-wise spike filter - removes data with a LHS differential exceeding 0.5
                filtered values replaced by nan values. 
                
                Args:
                    X: numpy array of echo intensity data with dimensions (nbins,n_ensembles)
                Returns:
                    X_new: numpy array with dimensions (nbins,n_ensembles)  
                    X_filt: filter mask 
                """
                X_norm = X - np.nanmin(X)
                X_norm = X_norm/np.nanmax(X_norm)
                
                X_dx = np.diff(X_norm,axis = 1) # difference across rows(up/down)
                X_dy = np.diff(X_norm,axis = 0) # difference across columns(left/right)
                
                X_new =X[:,1:].copy()
                X_filt = X_dx 
                X_new[X_filt>dx_thresh] = np.nan
                return X_new,X_filt
            
        
            base_mask = np.isnan(X) # mask for originalyl nan valeus
            
            X_new = X
            for _iter in range(niter):
                X_new,X_filt = filt(X_new) # filter out entries with large x-derivative
        
                mask = np.isnan(X_new)
                X_new[mask] = np.interp(np.flatnonzero(mask), np.flatnonzero(~mask), X_new[~mask]) # fill nan values with nearest neighbor
                
            
            
            #Append first two columns back into data (these were lost during the two passes of spike filter)
            Xout = X.copy()
            
            if sigma:
                Xout[:,niter:] = gaussian_filter(X_new, sigma = sigma)
            else:
                Xout[:,niter:] = X_new
        
            Xout[base_mask] = np.nan #(nan back out originally nan values)
        
            
            return Xout
    
    
 
        X = self._pd0.get_ensemble_array(field_name = 'ECHO INTENSITY') # raw echo data

        X_filt = X
        for i in range(X.shape[0]):#loop over each beam
            X_filt[i] = filter_array(X[i])
    
        self.append_to_ensembles(X_filt,title = 'ECHO INTENSITY')
        
        

    def mask_by_platform_z(self, min_z=0, max_z=1e6):
        """
        Apply a mask to ADCP data based on platform depth (Z-coordinate).
    
        Parameters
        ----------
        min_z : float, optional
            Minimum depth in meters to include in the mask. Defaults to 0.
        max_z : float, optional
            Maximum depth in meters to include in the mask. Defaults to 1e6.
    
        Notes
        -----
        This method creates a 1D mask based on the platform's Z-coordinate and broadcasts it
        to match the dimensions of the data arrays. It then applies this mask to all ensembles,
        effectively filtering out data collected outside the specified depth range.
        """
        mask_1d = (self._pd0.geometry.pose.pose.z >= min_z) & (self._pd0.geometry.pose.pose.z <= max_z)
        mask = np.broadcast_to(mask_1d[np.newaxis, np.newaxis, :], (self._pd0.config.number_of_beams, self._pd0.config.number_of_cells, self._pd0.n_ensembles))
        self._pd0.mask.define_mask(mask=mask, name='By Platform Depth', set_active=True)
    
    def mask_by_correlation_magnitude(self, min_cmag=64, max_cmag=256):
        """
        Apply a mask based on correlation magnitude values.
    
        Parameters
        ----------
        min_cmag : int, optional
            Minimum correlation magnitude to include in the mask. Defaults to 64.
        max_cmag : int, optional
            Maximum correlation magnitude to include in the mask. Defaults to 256.
    
        Notes
        -----
        Creates and applies a mask to ADCP ensemble data based on the correlation magnitude.
        Data with correlation magnitude values outside the specified range are masked out.
        """
        X = self._pd0.get_ensemble_array(field_name='CORRELATION MAGNITUDE')
        mask = (X >= min_cmag) & (X <= max_cmag)
        self._pd0.mask.define_mask(mask, name=f'CORRELATION MAGNITUDE [{min_cmag} ,{max_cmag}]', set_active=True)
    
        
    def mask_by_signal_to_noise(self, min_stn=1, max_stn=1e4):
        """
        Apply a mask based on signal to noise ratio.
    
        Parameters
        ----------
        min_stn : float, optional
            Minimum signal to noise ratio to include in the mask. Defaults to 1.
        max_stn : float, optional
            Maximum signal to noise ratio to include in the mask. Defaults to 1e4.
    
        Notes
        -----
        Creates and applies a mask to ADCP ensemble data based on the signal to noise ratio.
        Data with signal to noise ratio values outside the specified range are masked out.
        """
        X = self._pd0.get_ensemble_array(field_name='SIGNAL TO NOISE RATIO')
        mask = (X >= min_stn) & (X <= max_stn)
        self._pd0.mask.define_mask(mask, name=f'SIGNAL TO NOISE RATIO [{min_stn} ,{max_stn}]', set_active=True)
    
        
    def mask_by_absolute_backscatter(self, min_abs=-95, max_abs=0):
        """
        Apply a mask based on absolute backscatter values.
    
        Parameters
        ----------
        min_abs : int, optional
            Minimum absolute backscatter value to include in the mask. Defaults to -95 dB.
        max_abs : int, optional
            Maximum absolute backscatter value to include in the mask. Defaults to 0 dB.
    
        Notes
        -----
        Creates and applies a mask to ADCP ensemble data based on the absolute backscatter values.
        Data with absolute backscatter values outside the specified range are masked out.
        """
        X = self._pd0.get_ensemble_array(field_name='ABSOLUTE BACKSCATTER')
        mask = (X >= min_abs) & (X <= max_abs)
        self._pd0.mask.define_mask(mask, name=f'ABSOLUTE BACKSCATTER [{min_abs} ,{max_abs}]', set_active=True)
                
    def mask_bottom_track(self, cell_offset=0, spike_filter=False, spike_window=10, spike_thresh=5):
        """
        Create a mask based on the ADCP bottom track data, optionally applying a rolling spike filter.
    
        Parameters
        ----------
        cell_offset : int, optional
            The number of cells to mask above (positive) or below (negative) the bottom-track cell.
            Defaults to 0, which masks the bottom track cell itself.
        spike_filter : bool, optional
            Whether to apply a spike filtering to the bottom track data before creating the mask.
            Defaults to False.
        spike_window : int, optional
            The window size for the rolling spike filter. Only used if spike_filter is True.
            Defaults to 10.
        spike_thresh : int, optional
            The threshold factor for spike detection in the rolling spike filter. Only used if
            spike_filter is True. Defaults to 5.
    
        Notes
        -----
        This method calculates the ADCP bottom track range either directly or with a spike filter,
        then creates a mask that marks cells as being above or below the bottom track based on
        the specified offset. This mask is then applied to the ADCP data to filter out cells
        affected by proximity to the seabed or other reflective surfaces.
        """
        offset_distance = self._pd0.config.depth_cell_length * cell_offset
    
        if spike_filter:
            bt_range = self._pd0.get_bottom_track()
            
            for i in range(bt_range.shape[0]):
                bt_range[i] = ptools.rolling_spike_filter(bt_range[i] , window_size=spike_window, threshold_factor=spike_thresh)
        else:
            bt_range = self._pd0.get_bottom_track()
    
        # mask by pg
        for i in range(bt_range.shape[0]):
            mask = self._pd0.get_leader_data(leader = 'BOTTOM TRACK', field = f'BEAM#{i+1} BT %GOOD')>0
            
            bt_range[i][~mask] = np.nan# ptools.rolling_spike_filter(bt_range[i] , window_size=spike_window, threshold_factor=spike_thresh)
    
        
        if any(~np.isnan(bt_range).flatten()):
            bt_range = np.repeat(bt_range[:, np.newaxis, :], self._pd0.config.number_of_cells, axis=1)
            bt_range[bt_range == 0] = 32675  # Presumed missing data code
            bin_centers = np.outer(self._pd0.geometry.get_bin_midpoints(), np.ones(self._pd0.n_ensembles))
            bin_centers = np.repeat(bin_centers[np.newaxis, :, :], self._pd0.config.number_of_beams, axis=0)
            mask = (bt_range - self._pd0.config.depth_cell_length / 100) > (bin_centers + cell_offset)
            self._pd0.mask.define_mask(mask=mask, name='bottom track', set_active=True)

         
    def _scalar_counts_to_absolute_backscatter(self,E,E_r,k_c,alpha,C,R,Tx_T, Tx_PL, P_DBW):
        """
        Absolute Backscatter Equation from Deines (Updated - Mullison 2017 TRDI Application Note FSA031)
    
        Parameters
        ----------
        E_r : float
            Measured RSSI amplitude in the absence of any signal (noise), in counts.
        C : float
            Constant combining several parameters specific to each instrument.
        k_c : float
            Factor to convert amplitude counts to decibels (dB).
        E : float
            Measured Returned Signal Strength Indicator (RSSI) amplitude, in counts.
        Tx_T : float
            Tranducer temperature in deg C.
        R : float
            Along-beam range to the measurement in meters
        alpha : float
            Acoustic absorption (dB/m). 
        Tx_PL : float
            Transmit pulse length in dBm.
        P_DBW : float
            Transmit pulse power in dBW.
    
        Returns
        -------
        tuple
            Tuple containing two elements:
            - Sv : float
                Apparent volume scattering strength.
            - StN : float
                True signal to noise ratio.
    
        Notes
        -----
        - The use of the backscatter equation should be limited to ranges beyond π/4 * Rayleigh Distance for the given instrument.
        - Rayleigh Distance is calculated as transmit pulse length * α / width, representing the distance at which the beam can be considered to have fully formed.
        - For further details, refer to the original documentation by Deines.
        - 𝑘_c is a factor used to convert the amplitude counts reported by the ADCP’s receive circuitry to decibels (dB).
        - 𝐸 is the measured Returned Signal Strength Indicator (RSSI) amplitude reported by the ADCP for each bin along each beam, in counts.
        - 𝐸_r is the measured RSSI amplitude seen by the ADCP in the absence of any signal (the noise), in counts, and which is constant for a given ADCP.
        """
        StN = (10**(k_c * E / 10) - 10**(k_c * E_r / 10)) / (10**(k_c * E_r / 10))
        L_DBM = 10 * np.log10(Tx_PL)
        #P_DBW = 10 * np.log10(Tx_Pw)
        Sv = C + 10 * np.log10((Tx_T + 273.16) * (R**2)) - L_DBM - P_DBW + 2 * alpha * R + 10 * np.log10((10**(0.1 * k_c * (E - E_r)) - 1))
    
        return Sv, StN
    
       
        
    def _scalar_water_absorption_coeff(self,T, S, z, f, pH):
        '''
        Calculate water absorption coefficient.
    
        Parameters
        ----------
        T : float
            Temperature in degrees Celsius.
        S : float
            Salinity in practical salinity units (psu).
        z : float
            Depth in meters.
        f : float
            Frequency in kHz.
        pH : float
            Acidity.
    
        Returns
        -------
        float
            Water absorption coefficient in dB/km.
        '''
        c = 1449.2 + 4.6 * T - 0.055 * T**2 + 0.00029 * T**3 + (0.0134 * T) * (S - 35) + 0.016 * z
        #c = 1412 + 3.21 * T + 1.19 * S + 0.0167 * z
    
        # Boric acid component
        A1 = (8.68 / c) * 10**(0.78 * pH - 5)
        P1 = 1
        f1 = 2.8 * ((S / 35)**0.5) * 10**(4 - (1245 / (273 + T)))
    
        # Magnesium sulphate component
        A2 = 21.44 * (S / c) * (1 + 0.025 * T)
        P2 = 1 - (1.37e-4) * z + (6.2e-9) * z**2
        f2 = (8.17 * 10**(8 - (1990 / (273 + T)))) / (1 + 0.0018 * (S - 35))
    
        if T <= 20:
            A3 = (4.937e-4) - (2.59e-5) * T + (9.11e-7) * T**2 - (1.5e-8) * T**3
        elif T > 20:
            A3 = (3.964e-4) - (1.146e-5) * T + (1.45e-7) * T**2 - (6.5e-8) * T**3
        P3 = 1 - (3.83e-5) * z + (4.9e-10) * (z**2)
    
        # Calculate water absorption coefficient
        alpha_w = (A1 * P1 * f1 * (f**2) / (f**2 + f1**2) + A2 * P2 * f2 * (f**2) / (f**2 + f2**2) + A3 * P3 * (f**2))
        
        # Convert absorption coefficient to dB/km
        alpha_w = (1 / 1000) * alpha_w
        
        return alpha_w
    
    
    def _scalar_sediment_absorption_coeff(self,ps, pw, d, SSC, T,S, f,z):
        '''
        Calculate sediment absorption coefficient.
    
        Parameters
        ----------
        ps : float
            Particle density in kg/m^3.
        pw : float
            Water density in kg/m^3.
        d : float
            Particle diameter in meters.
        SSC : float
            Suspended sediment concentration in kg/m^3.
        T : float
            Temperature in degrees Celsius.
        f : float
            Frequency in kHz .
    
        Returns
        -------
        float
            Sediment absorption coefficient.
        '''
        c = 1449.2 + 4.6 * T - 0.055 * T**2 + 0.00029 * T**3 + (0.0134 * T) * (S - 35) + 0.016 * z # speed of sound in water
        v = (40e-6) / (20 + T)  # Kinematic viscosity (m2/s)
        B = (np.pi * f / v) * 0.5
        delt = 0.5 * (1 + 9 / (B * d))
        sig = ps / pw
        s = 9 / (2 * B * d) * (1 + (2 / (B * d)))
        k = 2 * np.pi / c  # Wave number (Assumed, as it isn't defined in the paper)
    
        alpha_s = (k**4) * (d**3) / (96 * ps) + k * ((sig - 1)**2) / (2 * ps) + \
                  (s / (s**2 + (sig + delt)**2)) * (20 / np.log(10)) * SSC
    
        return alpha_s     

        

    def calculate_absolute_backscatter(self,**kwargs):
        """
        Convert numpy array of ensemble data of echo intensity to absolute backscatter then added to ensemble_data property of the WorkhorseADCP class.
    
        Optional Args:
            E_r
            k_c
            C
            alpha 
            P_dbw
     
        """
    
    
        E_r = 39 # noise floor 
        
        ## select C based on WB commmand (0 = 25%, 1 = 6.25%)
        WB = self._pd0.ensemble_data[0]['FIXED LEADER']['SYSTEM BANDWIDTH {WB}']
        if WB==0:
            C = -139.09 #for Workhorse 600, 25% 
        else:
            C = -149.14 # for Workhorse 600, 6%
        # Beam High Gain RSSI
        
        if hasattr(self._pd0,'PT3'):
            k_c = self._pd0.PT3.k_c
        else:
            k_c = {1: 0.3931,# beam 1
                    2: 0.4145,# beam 2
                    3: 0.416,# beam3
                    4: 0.4129}# beam4
        if self._pd0.ensemble_data[0]["SYSTEM CONFIGURATION"]["FREQUENCY"] == '300-kHz':
            alpha = 0.068 #nominal ocean value for a 600 kHz unit 
            P_dbw = 14 #battery supply power for Workhorse 300
            print('300 kHz Unit')
        elif self._pd0.ensemble_data[0]["SYSTEM CONFIGURATION"]["FREQUENCY"] == '600-kHz':
            alpha =  0.178 #nominal ocean value for a 600 kHz unit 
            P_dbw = 9 #battery supply power for Workhorse 600
            
        elif self._pd0.ensemble_data[0]["SYSTEM CONFIGURATION"]["FREQUENCY"] == '75-kHz':
            alpha =  0.027 #nominal ocean value for a 600 kHz unit 
            P_dbw = 27.3 #battery supply power for Workhorse 600
            
            
        # overwrite defaults with kwarg parameters 
        E_r = kwargs.pop('E_r',E_r) 
        C = kwargs.pop('C',C)
        k_c = kwargs.pop('k_c',k_c)
        alpha = kwargs.pop('alpha',alpha)
        P_dbw = kwargs.pop('P_dbw',P_dbw)
        
    
        #get other instrument data
        temperature = np.outer(self._pd0.get_sensor_temperature(),np.ones(self._pd0.config.number_of_cells)).T
        bin_distances = np.outer(self._pd0.geometry.get_bin_midpoints(),np.ones(self._pd0.n_ensembles))#/np.cos(20*np.pi/180)#/100
        transmit_pulse_lengths = np.outer(self._pd0.get_sensor_transmit_pulse_lengths(),np.ones(self._pd0.config.number_of_cells)).T#/100
        E = self._pd0.get_ensemble_array(field_name = 'ECHO INTENSITY') #echo intensity array 
        
        # initalize arrays  
        X = np.empty((self._pd0.config.number_of_beams,self._pd0.config.number_of_cells,self._pd0.n_ensembles))
        StN = np.empty((self._pd0.config.number_of_beams,self._pd0.config.number_of_cells,self._pd0.n_ensembles))
    
        for beam in range(self._pd0.config.number_of_beams):
            # StN[beam,:,:] = (10**(k_c[beam+1]*E[beam,:,:]/10) - 10**(k_c[beam+1]*E_r/10))/10**(k_c[beam+1]*E_r/10)
            # X[beam,:,:] = C + 10*np.log10((temperature + 273.16)*(bin_distances**2)) - 10*np.log10(transmit_pulse_lengths) - P_dbw + 2*alpha*bin_distances + 10*np.log10(10**(k_c[beam+1]*(E[beam,:,:] - E_r)/10)-1)
            
            sv,stn =self.counts_to_absolute_backscatter(E = E[beam],
                                                        E_r = E_r,
                                                        k_c= k_c[beam+1],
                                                        alpha = alpha,
                                                        C = C,
                                                        R = bin_distances,
                                                        Tx_T = temperature,
                                                        Tx_PL = transmit_pulse_lengths,
                                                        P_DBW = P_dbw)

            
            StN[beam] = stn
            X[beam] = sv
            
            
            #X[beam,:,:] = C + 10*np.log10((temperature*bin_distances**2)/(transmit_pulse_lengths*P_dbw)) + 2*alpha*bin_distances + k_c[beam+1]*(E[beam,:,:] - E_r)
        self.append_to_ensembles(X,'ABSOLUTE BACKSCATTER')
        self.append_to_ensembles(StN,'SIGNAL TO NOISE RATIO')   
        
        
        
        
    def append_to_ensembles(self,X,title,nan_value = 32768):
        """
        Format numpy array of ensemble data into a list then add to ensemble_data property of the WorkhorseADCP class.
        Input array must have dimensions of (n_beams,nbins,n_ensemles). In a typical use case, ensemble data is manipulated in 
        array format, then appended to the ensemble_data property and written to a PD0 file. 
        Args:
            X: numpy array with dimensions (n_bins,n_ensembles,n_beams)
            title: (string) title for the formatted data when appended to the ensemble_data property of the WorkhorseADCP class object. 
            nan_value: (int) value to use in place of nan. 32768 is default. 
        """
        self.ensemble_fields.append(title) # update the list of ensemble field names
        for e in range(self._pd0.n_ensembles):
            ensemble_data = [] # data in the current ensemble
            for b in range(self._pd0.config.number_of_cells):
                bin_data = [] # data in the current bin
                for bm in range(self._pd0.config.number_of_beams):
                    
                    val = X[bm,b,e]
                    #bin_data.append(int(val))
                    try:
                        bin_data.append(int(val))
                    except:
                        bin_data.append(nan_value)
                        
                        
                ensemble_data.append(bin_data)
            self._pd0.ensemble_data[e][title] = ensemble_data  